import httplib
import json
import uuid

from furl import furl
import pytest
from pytest_localserver.http import WSGIServer
from werkzeug import Request
from werkzeug import Response


class FestungApp(WSGIServer):
    def __init__(self, *args, **kwargs):
        kwargs.update(application=self)
        super(FestungApp, self).__init__(*args, **kwargs)
        self.requests = []
        self.responses = []

    @Request.application
    def __call__(self, request):
        try:
            response = self.responses[len(self.requests)]
        except IndexError:
            response = Response("No prepared response")
            response.status_code = httplib.NOT_IMPLEMENTED
            return response

        self.requests.append(request)
        return response

    def add_response(self, data, status=None, headers=None):
        response = Response(data)
        response.status_code = status or httplib.OK
        response.headers = headers or {'Content-Type': 'text/html; charset=UTF-8'}
        self.responses.append(response)

    def add_json_response(self, data, status=httplib.OK):
        headers = {'Content-Type': 'application/json'}
        self.add_response(json.dumps(data), status=status, headers=headers)

    def add_empty_response(self):
        self.add_response(None, status=httplib.NO_CONTENT)

    @property
    def json_requests(self):
        return [json.loads(r.get_data()) for r in self.requests]


@pytest.fixture
def festung():
    server = FestungApp()
    server.start()
    try:
        yield server
    finally:
        server.stop()


@pytest.fixture
def festung_url(festung):
    return furl(festung.url).set(scheme='festung').url


@pytest.fixture
def database():
    return uuid.uuid4().hex


@pytest.fixture
def password():
    return uuid.uuid4().hex


@pytest.fixture(params=[''])
def query_string(request):
    return request.param


@pytest.fixture
def database_url(festung_url, password, database, query_string):
    url = furl(festung_url)
    assert url.username is None and url.password is None
    return url.set(password=password, path=database, query=query_string).remove(fragment=True).url



from __future__ import absolute_import

import exceptions


class Error(exceptions.StandardError):

    def __init__(self, message, error=None):
        super(Error, self).__init__(message)
        self.error = error


class Warning(exceptions.StandardError):
    pass


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class InternalError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class DataError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass



import base64
import datetime
import decimal
import urllib
import warnings

from furl import furl


# Protocol used by festing in reality. This has to be changed
# to HTTPS at some point.
PROTO = 'http'


def no_password_url(url):
    """Strip the password from the url.

    Example:

        >>> no_password_url('festung://login:password@host:8080/path')
        'festung://login:********@host:8080/path'
        >>> no_password_url('festung://login@[2001:db8::2]/path')
        'festung://login@[2001:db8::2]/path'
    """
    url = furl(url)
    if url.password is not None:
        url.password = '********'  # We don't want to leak the length of the password
    return urllib.unquote(url.url)


def to_http_url(url):
    """Convert a ``festung://`` url to the ``http`` url.

    Example:
        >>> to_http_url('festung://:password@example.com:2827/vault')
        'http://example.com:2827/vault'
        >>> to_http_url('festung://:password@[2001:db8::2]:2827/path')
        'http://[2001:db8::2]:2827/path'
        >>> to_http_url('festung://example.com:2324/vault?kdf-iter=4000')
        'http://example.com:2324/vault'
    """
    return furl(url).set(scheme=PROTO).remove(username=True, password=True, query=True).url


def to_headers(url):
    """Export HTTP headers from the `url`.

    This function extracts the 'Authorization' header and other paramters that are given in the
    query string and returns a dict of headers.

    Example::
        >>> to_headers('festung://:password@example.com:2827/vault?kdf-iter=4000') == {
        ...     'X-kdf-iter': '4000',
        ...     'Authorization': 'cGFzc3dvcmQ='}
        True
        >>> to_headers('festung://:passwd@server.tld/path?a=1') == {'Authorization': 'cGFzc3dk'}
        True
        >>> to_headers('festung://server.tld/path?b=1') == {}
        True
    """
    url = furl(url)
    headers = {}
    if url.password:
        headers['Authorization'] = base64.b64encode(url.password)
    headers.update(qs_to_headers(url.url))
    return headers


PARAM_TO_HEADER = {
    'kdf-iter': 'X-kdf-iter'
}


def qs_to_headers(url):
    return {PARAM_TO_HEADER[k]: v for k, v in furl(url).args.items() if k in PARAM_TO_HEADER}


def cast(obj):
    """Cast most python types to valid types to be serialized in JSON and passed to SQLite"""
    if isinstance(obj, bool):
        # checking bool first as it is also an instance of int
        return int(obj)
    if obj is None or isinstance(obj, (long, int, float, basestring)):
        if isinstance(obj, float):
            warnings.warn(
                "floats are highly inaccurate. Use them at your own risk.", RuntimeWarning)
        return obj

    if isinstance(obj, decimal.Decimal):
        return str(obj)

    if isinstance(obj, datetime.datetime):
        if obj.tzinfo is None:
            warnings.warn("non-timezone aware datetime can lead to many issues.", RuntimeWarning)
        return obj.isoformat()

    if isinstance(obj, (datetime.date, datetime.time)):
        return obj.isoformat()

    raise TypeError("Unserializeable type {}".format(type(obj)))



import requests


class SessionPool(object):
    """Pool to manage connections to festung.

    This pool expects the session connection to be managed externally. It
    doesn't create nor close the TCP connection to festung.

    Args:
        session (requests.Session): A pre-created, externally managed `Session`
            from the requests library.

    Attributes:
        session (requests.Session): The requests' session that this pool use to
            connect to festung.
    """

    def __init__(self, session):
        self.session = session

    def request(self, request):
        resp = self.session.send(request.prepare())
        resp.raise_for_status()
        return resp

    def close(self):
        pass


class NewSessionPool(SessionPool):
    """Same as :class:`SessionPool` but instanciate a new `requests.Session`.

    Args:
        session (NoneType): Only accept None as a session (for polymorphism)
    """
    def __init__(self, session):
        if session is not None:
            raise TypeError("session should be None")
        super(NewSessionPool, self).__init__(requests.Session())

    def close(self):
        super(NewSessionPool, self).close()
        self.session.close()


def get_pool_class(session):
    if session is None:
        return NewSessionPool
    else:
        return SessionPool


def get_pool(session):
    pool_class = get_pool_class(session)
    return pool_class(session)
